﻿using Buddy.Coroutines;
using Clio.Utilities;
using Clio.XmlEngine;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Objects;
using ff14bot.RemoteWindows;
using System.Linq;
using System.Threading.Tasks;
using TreeSharp;
using Action = TreeSharp.Action;

namespace ff14bot.NeoProfiles {
    [XmlElement("NPCRepair")]
    public class NPCRepairTag : ProfileBehavior {
        [XmlAttribute("NpcId")]
        public uint NpcId { get; set; }

        [XmlAttribute("XYZ")]
        public Vector3 XYZ { get; set; }

        [XmlAttribute("Threshhold")]
        public float Threshhold { get; set; }
        
        private GameObject NPC;

        private bool _done = false;

        private bool RepairAllClicked = false;
        private bool Repaired = false;

        public override bool IsDone { get { return _done; } }

        protected Composite Behavior() 
		{
            return 
				new PrioritySelector(
					// can tag execute?
					new Decorator(
						ret => !CanRepair(),
						new Action( 
							r=> 
							{
								_done = true;
							})),
					// Find the NPC
					new Decorator(
						ret => NPC == null,
						new Action(
							r => 
							{
								NPC = GameObjectManager.GetObjectByNPCId(NpcId);
								
								if (NPC == null)
								{
									_done = true;
								}
								else
								{
									Poi.Current = new Poi(NPC,PoiType.Vendor);
								}
							})),
					// go to NPC and Interact
					new Decorator(
						ret => !Repair.IsOpen,
						new PrioritySelector(
							new Decorator(
								ret => Vector3.Distance(Core.Player.Location, XYZ) > 1.0,
								CommonBehaviors.MoveTo(
									ret => XYZ,
									"Mender")),
							new Sequence(
								CommonBehaviors.MoveStop(),
								new Sleep(1000),
								new Action(
									r => 
									{
										Log ("Interacting with the Mender");
																				
										Poi.Current.Unit.Interact();
									}),
								new Sleep(1000)))),
					// repair
					new Decorator(
						ret => Repair.IsOpen,
						new PrioritySelector(
							new Decorator(
								ret => Repaired,
								new Action(
									r => 
									{
										Repair.Close();
										_done = true;
									})),
							new Decorator(
								ret => Repair.IsOpen && ! RepairAllClicked,
								new Action(
									r => 
									{
										RepairAllClicked = true;
										Repair.RepairAll();
									})),
							new Decorator(
								ret => SelectYesno.IsOpen,
								new Action(
									r => 
									{
										SelectYesno.ClickYes();
										Repaired = true;
									})))));
	}

        public bool CanRepair() {
            if (Threshhold == null || Threshhold <= 0 || Threshhold > 100)
                Threshhold = 100f;

            return InventoryManager.EquippedItems.Where(r => r.IsFilled && r.Condition < Threshhold).Count() > 0;
        }
		
		protected void Initialize()
		{
			_done = false;
			Repaired = false;
			RepairAllClicked = false;
			NPC = null;
			Poi.Clear("RepairTag Initialized");
		}

        protected override void OnResetCachedDone() 
		{
			Initialize();
        }

        protected override void OnDone() 
		{
            TreeHooks.Instance.RemoveHook("TreeStart", _cache);
        }

        private Composite _cache;

        protected override void OnStart() 
		{
			Initialize();
			
            _cache = Behavior();
            TreeHooks.Instance.InsertHook("TreeStart", 0, _cache);
        }
    }
}