using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Media;
using System.Threading;
using System.Threading.Tasks;

using Buddy.Coroutines;

using Clio.Common;
using Clio.Utilities;
using Clio.XmlEngine;

using ff14bot;
using ff14bot.AClasses;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Objects;
using ff14bot.RemoteWindows;
using ff14bot.Enums;
using ff14bot.NeoProfiles;
using ff14bot.Settings;

using TreeSharp;
using Action = TreeSharp.Action;

namespace ff14bot.NeoProfiles {
    [XmlElement("TurnInCollectable")]
    public class TurnInCollectableTag : ProfileBehavior 
	{
        [XmlAttribute("NpcId")]
        public uint NpcId { get; set; }

        [XmlAttribute("XYZ")]
        public Vector3 XYZ { get; set; }
		
		[XmlAttribute("Collectable")]
		public string Collectable { get; set; }
        
		[XmlAttribute("AppraiserWindowLineNumber")]
		public uint AppraiserWindowLineNumber { get; set; }
        
        private const uint WM_KEYDOWN  = 0x100;
        private const uint WM_KEYUP    = 0x0101;
		private const int  KEY_Numpad0 = 0x60;
		private const int  KEY_Numpad2 = 0x62;
		private const int  KEY_Decimal = 0x6E;
		private const int  KEY_Prior   = 0x21;
		
        private BotEvent cleanup;
        private GameObject NPC;
        private Composite _cache;
        private bool _done = false;
		private bool isAppraiserWindowOpen = false;
		private bool isCollectableSelectedForTurnIn = false;
		private bool isCollectableRejected = false;

        [return: MarshalAs(UnmanagedType.Bool)]
        [DllImport("user32.dll")]
        protected static extern bool PostMessage(IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam);
		
		protected void Initialize()
		{
            _done = false;
			isAppraiserWindowOpen = false;
			isCollectableRejected = false;
			isCollectableSelectedForTurnIn = false;
			NPC = null;
			Poi.Clear("TurnInCollectableTag Initialized");
		}

        protected void ReceiveMessage(object sender, ChatEventArgs e)
        {
            if (e.ChatLogEntry.MessageType == MessageType.SystemMessages &&
                e.ChatLogEntry.Contents.StartsWith("Unable to complete delivery"))
            {
				Log("The Collectable Appraiser is not accepting the item at line "+AppraiserWindowLineNumber);
				
				isCollectableRejected = true;
			}
		}
		
        public override bool 
		IsDone 
		{ 
			get 
			{ 
				return _done; 
			} 
		}
		
        protected override void OnStart() 
		{
			Initialize();
			
            GamelogManager.MessageRecevied += ReceiveMessage;
			
            cleanup = 
				(BotBase bot) =>
                {
					try
					{
						GamelogManager.MessageRecevied -= ReceiveMessage;
					}
					catch (Exception ex)
					{
						Log(ex.Message);
					}
					
					TreeHooks.Instance.RemoveHook(
						"TreeStart", 
						_cache);
						
					TreeRoot.OnStop -= cleanup;
                };

            _cache = Behavior();
		   
            TreeHooks.Instance.InsertHook(
				"TreeStart", 
				0, 
				_cache);
            TreeRoot.OnStop += cleanup;
        }
		
        protected override void OnResetCachedDone() 
		{
			Initialize();
        }

        protected override void OnDone() 
		{
			cleanup(
				null);
        }

        protected async Task<bool> PostKeyPress(int key)
        {
            PostMessage(Core.Memory.Process.MainWindowHandle, WM_KEYDOWN, new IntPtr(key), IntPtr.Zero);
            PostMessage(Core.Memory.Process.MainWindowHandle, WM_KEYUP, new IntPtr(key), IntPtr.Zero);
			await Coroutine.Sleep(500);
			
			return true;
        }
		
		protected async Task<bool>  CloseAppraiserWindow()
		{
			if (isAppraiserWindowOpen)
			{
				await PostKeyPress(KEY_Decimal);
				
				isAppraiserWindowOpen = false;
			}
			
			return true;
		}
		
		public static BagSlot GetItem(string itemName)
		{
			try
			{
				return
					InventoryManager.FilledSlots.First(
						s => s.EnglishName == itemName && s.IsCollectable);
			}
			catch (InvalidOperationException ex)
			{
			}
			
			return null;
		}
		
		public static int CountItem(string itemName)
		{
			return
				InventoryManager.FilledSlots.Count(
					s => s.EnglishName == itemName && s.IsCollectable);
		}
		
		public static bool HaveItem(string itemName)
		{
			return
				CountItem(itemName) > 0;
		}

		protected bool HaveCollectable
		{
			get
			{
				return
					!string.IsNullOrEmpty(Collectable) &&
					HaveItem(Collectable);
			}
		}
				
		public async Task<bool> TurnInCollectable() 
		{
			if (!isCollectableSelectedForTurnIn)
			{
				Log("Selecting Collectable");
				
				await PostKeyPress(KEY_Prior);
				
				for (uint i = 1;
					 i <= AppraiserWindowLineNumber-1;
					 i++)
				 {
					await PostKeyPress(KEY_Numpad2);
				 }
					
				isCollectableSelectedForTurnIn = true;         
			}
			else
			{
				// Need to bring appraiser window back into focus after a hand over
				
				await PostKeyPress(
					KEY_Numpad0);
			}
			
			Log("Activating Collectable");
			
			int items = CountItem(Collectable);
				
			await PostKeyPress(
				KEY_Numpad0);
				
			await Coroutine.Wait(
				3000,
				() => Request.IsOpen);
			
			Log("Retrieving Collectable to hand over");
			
			BagSlot collectable = GetItem(Collectable);
						
			collectable.Handover();
				
			await Coroutine.Sleep(1000);
			
			Log("Handing over collectable");
			
			Request.HandOver();
			
			await Coroutine.Sleep(1000);
			
			if (CountItem(Collectable) == items)
			{
				Log("The Collectable Appraiser is not accepting the "+Collectable+" at line "+AppraiserWindowLineNumber);
				
				isCollectableRejected = true;
			}
			
            return true;
        }
		
		protected Composite Behavior()
 		{
            return 
				new PrioritySelector(
					// Make sure we have the specified collectable and have been told where it is in the appraiser window
					new Decorator(
						ret => _done ||
							   !HaveCollectable || 
							   AppraiserWindowLineNumber == 0 || 
							   NpcId == 0 ||
							   XYZ == null,
						new ActionRunCoroutine(
							async(r) =>
							{
								await CloseAppraiserWindow();
								
								_done = true;
							})),
					// Find the NPC
					new Decorator(
						ret => NPC == null,
						new Action(
							r => 
							{
								NPC = GameObjectManager.GetObjectByNPCId(NpcId);
								
								if (NPC == null)
								{
									_done = true;
								}
								else
								{
									Poi.Current = new Poi(NPC,PoiType.Vendor);
								}
							})),
					// If yes/no dialog is up, we have have overrun our turn in quota
					new Decorator(
						ret => SelectYesno.IsOpen,
						new ActionRunCoroutine(
							async(r) =>
							{
								SelectYesno.ClickNo();
								
								//Need to bring the appraiser window back into focus after dismissing yes/no
								
								await PostKeyPress(
									KEY_Numpad0);
								
								await CloseAppraiserWindow();
								
								_done = true;
							})),
					// Make sure the Collectable Appraiser wants what we are offering
					new Decorator(
						ret => isCollectableRejected,
						new ActionRunCoroutine(
							async(r) =>
							{
								await CloseAppraiserWindow();
								
								_done = true;
							})),
					// Go to NPC and Interact
					new Decorator(
						ret => !isAppraiserWindowOpen,
						new PrioritySelector(
							new Decorator(
								ret => Vector3.Distance(Core.Player.Location, XYZ) > 1.0,
								CommonBehaviors.MoveTo(
									ret => XYZ,
									"Collector Appraiser")),
							new Sequence(
								CommonBehaviors.MoveStop(),
								new ActionRunCoroutine(
									async(r) => 
									{
										Log ("Interacting with the Collectable Appraiser");
										
										Poi.Current.Unit.Interact();
										
										await Coroutine.Sleep(1000);
										
										isAppraiserWindowOpen = true;
									})))),
					// Turn in collectables
					new ActionRunCoroutine(					
						r => TurnInCollectable()));
        }
    }
}